// MgaProject.cpp : Implementation of CMgaProject
#include "stdafx.h"
#include "CommonComponent.h"
#include "MgaFCO.h"
#include "MgaFilter.h"
#include "stdio.h"

void LockingStart(CComPtr<ICoreProject> &tempproject)	{ 
		   CComPtr<ICoreTerritory> lm;
		   COMTHROW(tempproject->CreateTerritory( &lm));
		   COMTHROW(tempproject->PushTerritory( lm));
}

/////////////////////////////////////////////////////////////////////////////
// CMgaProject
/////////////////////////////////////////////////////////////////////////////

CMgaProject::CMgaProject()	{ 
			INITSIG('P');
#ifdef DEBUG
			MGA_TRACE("Constructed: %s - %08X\n", sig, this);
#endif
			opened = CLOSED;
			dataproject = NULL;
			preferences = 0;
			opmask = 0x0026662A;
			mgaproject = this;
			rot = 0;
			checkoff = false;
			transactioncount = 0;
			autoaddons = false;
			inautoaddoncreate = false;
			mgaversion = 0;
			aurcnt = 0;
			guidstat = CLEAN;
}

STDMETHODIMP CMgaProject::FinalConstruct()	{
		COMTRY {
//			IDispatchImpl<IMgaProject, &IID_IMgaProject, &LIBID_MGALib> *xthis = this;
			COMTHROW( CoCreateInstance(CLSID_CoreProject, (IMgaProject *)this, CLSCTX_ALL, IID_IUnknown, (LPVOID *) &inner) );
			COMTHROW(inner.QueryInterface(&dataproject));
			dataproject->Release();   // release reference to ourselves
		} COMCATCH(;);
}

CMgaProject::~CMgaProject()	{
#ifdef DEBUG
		MGA_TRACE("Destructed: %s - %08X\n", sig, this);
#endif
		MARKSIG('9'); 
		ASSERT(allterrs.empty());
		if(opened != CLOSED) { 
			Close();
		}
}

/////////////////////////////////////////////////////////////////////////////
///////////////////////// META CACHE FACILITY ///////////////////////////////
/////////////////////////////////////////////////////////////////////////////

STDMETHODIMP CMgaProject::get_MetaObj(metaref_type id, IMgaMetaBase **pVal) {
	COMTRY {
		COMTHROW(metapr != 0 ? S_OK : E_MGA_PROJECT_NOT_OPEN);
		COMTHROW(metapr->get_FindObject(id, pVal));
	}
	COMCATCH(;);	
}

// THROWS
CComPtr<IMgaMetaBase> CMgaProject::FindMetaRef(metaref_type l) {
		CComPtr<IMgaMetaBase> hh;
		COMTHROW(metapr != 0 ? S_OK : E_MGA_PROJECT_NOT_OPEN);
		COMTHROW(metapr->get_FindObject(l, &hh));
//		HRESULT hr = metapr->get_FindObject(l, &hh);
//		if(hr != S_OK && hr != E_NOTFOUND) COMTHROW(hr);
		return hh;
}


///////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////    CREATE/OPEN PROJECT   //////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////


HRESULT CMgaProject::CreateSetupData(BSTR rootname, BSTR paradigmname, VARIANT guid) {
		COMTRY_IN_TRANSACTION {	

// check if meta is valid
			CoreObj  dataroot, rootfolder;
			CComBSTR parversion;

// set values in 			
			COMTHROW(dataproject->get_RootObject(&dataroot.ComPtr()));
			dataroot[ATTRID_CDATE] = dataroot[ATTRID_MDATE]  = Now();
			dataroot[ATTRID_PARADIGM] = paradigmname;
			dataroot[ATTRID_MGAVERSION] = mgaversion;
			COMTHROW(metapr->get_Version(&parversion));
			dataroot[ATTRID_PARVERSION] = parversion;
			dataroot[ATTRID_NAME]    = rootname;
			if( guid.vt != (VT_UI1 | VT_ARRAY) || GetArrayLength(guid) != sizeof(GUID) )
					COMTHROW(E_INVALIDARG);
			dataroot[ATTRID_PARGUID] = guid;
			dataroot[ATTRID_VERSION] = CComBSTR(""); // default version string

// create data root things
		    COMTHROW(dataproject->CreateObject(DTID_FOLDER,&rootfolder.ComPtr()));
			rootfolder[ATTRID_NAME]    = rootname;
			rootfolder[ATTRID_FPARENT] = dataroot;
			if(!(preferences & MGAPREF_MANUAL_RELIDS))  rootfolder[ATTRID_RELID] = 1;
			CComPtr<IMgaMetaFolder> mf;
			metaref_type mr;
			COMTHROW(metapr->get_RootFolder(&mf));
			COMTHROW(mf->get_MetaRef(&mr));
			rootfolder[ATTRID_META] = mr;
			notifyqueueprocessed = true;
		} COMCATCH_IN_TRANSACTION(;);

}


static int guidcmp(VARIANT &qGUID, VARIANT &pGUID)	{
	GUID g1, g2;
	CopyTo(qGUID, g1);
	CopyTo(pGUID, g2);
	return memcmp(&g1, &g2, sizeof(g1));
}


inline HRESULT EMAP(HRESULT hr, HRESULT from, HRESULT to) { HRESULT h = hr; if(h == from) h = to; return h; }
inline HRESULT EDEF(HRESULT hr, HRESULT to) { HRESULT h = hr; if(h != S_OK) h = to; return h; }


STDMETHODIMP CMgaProject::OpenParadigm(BSTR s, VARIANT *pGUID) {
	COMTRY {
		CComPtr<IMgaRegistrar> mgareg;
		COMTHROW(mgareg.CoCreateInstance(OLESTR("MGA.MgaRegistrar")));
		CComBSTR connstr;
		COMTHROW(EDEF(mgareg->QueryParadigm(s, &connstr, pGUID, REGACCESS_PRIORITY), E_MGA_PARADIGM_NOTREG));
		ASSERT(connstr);
		COMTHROW(metapr.CoCreateInstance(	OLESTR("MGA.MgaMetaProject")));
		COMTHROW(EDEF(metapr->Open(connstr), E_MGA_PARADIGM_NOTREG));
		CComVariant metaGUID;
		COMTHROW(metapr->get_GUID(&metaGUID));
		if(guidcmp(metaGUID, *pGUID)) COMTHROW(E_MGA_PARADIGM_INVALID);
		parconn = connstr;
	} COMCATCH(;);
}


STDMETHODIMP CMgaProject::OpenParadigm(BSTR s, BSTR ver) {
	COMTRY {
		CComVariant vguid;
		{
			CComPtr<IMgaRegistrar> mgareg;
			COMTHROW(mgareg.CoCreateInstance(OLESTR("MGA.MgaRegistrar")));
			COMTHROW(EDEF(mgareg->GUIDFromVersion(s, ver, &vguid, REGACCESS_PRIORITY), E_MGA_PARADIGM_NOTREG));
		}
		COMTHROW(OpenParadigm(s,&vguid));
	} COMCATCH(;);
}

STDMETHODIMP CMgaProject::CreateEx(BSTR projectname, BSTR paradigmname, VARIANT paradigmGUID) {
	COMTRY {
		if(opened != CLOSED) COMTHROW(E_MGA_PROJECT_OPEN);

		mgaversion = 1;
			// Set generic for meta
		CComPtr<ICoreMetaProject> genericproject;
		CreateCoreMetaProject(genericproject);

		CComVariant connGUID;
		if(paradigmGUID.vt != VT_EMPTY) connGUID = paradigmGUID;
		COMTHROW(OpenParadigm(paradigmname, &connGUID));
		COMTHROW(dataproject->CreateProject(projectname, genericproject));
		opened = UNCHANGED;
		guidstat = DIRTY;

	    CComPtr<IMgaTerritory> lm;
	    COMTHROW(CreateTerritory(NULL, &lm));
		COMTHROW(BeginTransaction(lm, TRANSACTION_GENERAL));
		
		HRESULT suhr = S_OK;
		try {
			CComBSTR   fname("Root Folder");
            // BGY: commented out because xml backend has different connection string format
			/*LPCOLESTR p = wcsrchr(projectname, '\\');
			if(p) {
				p++;
				LPCOLESTR p2 = wcschr(p, '.');
				if(p2 == NULL) p2 = p + wcslen(p);
				fname = CComBSTR(p2-p, p);
			}*/
			suhr = CreateSetupData(fname, paradigmname, connGUID);
			COMTHROW(lm->Flush());
			COMTHROW(CommitTransaction());
		} catch(hresult_exception &e) {
			lm->Flush();
			AbortTransaction();
			throw e;
		}
		COMTHROW(dataproject->FlushUndoQueue());
		COMTHROW(suhr);
		COMTHROW(RegisterActiveObject((IMgaProject *)this,CLSID_MgaProject,ACTIVEOBJECT_STRONG,&rot));
		projconn = projectname;
		StartAutoAddOns();
		try {
			COMTHROW(BeginTransaction(lm, TRANSACTION_READ_ONLY));
			GlobalNotify(GLOBALEVENT_OPEN_PROJECT);
			COMTHROW(lm->Flush());
			COMTHROW(CommitTransaction());
		} catch(hresult_exception &e) {
			lm->Flush();
			AbortTransaction();
			throw e;
		}
		MARKSIG('2');
	}
	COMCATCH(
		opened = CLOSED;
		guidstat = CLEAN;
		if(dataproject) dataproject->CloseProject();
		if(metapr) metapr->Close();
		metapr = 0;
		projconn.Empty(); 
		parconn.Empty();
	)
}


// if paradigname != "": 
//		change paradigmname if different
//		if paradigmguid == 0 update to the current version (even if no name change!);
// if paradigmguid != null and != oldguid update guid (use new paradigmname to locate paradigm in any case)
// if no guid change skip check
STDMETHODIMP CMgaProject::OpenEx(BSTR projectname, BSTR paradigmname, VARIANT paradigmGUID) {
	COMTRY {
		if(opened != CLOSED) COMTHROW(E_MGA_PROJECT_OPEN);
		CComBSTR s;
		CComVariant pGUID;
		CComBSTR ver;
		CComPtr<ICoreMetaProject> genericproject;
		CreateCoreMetaProject(genericproject);
		VARIANT_BOOL ro;

		COMTHROW(dataproject->OpenProject(projectname, genericproject, &ro));
		
		projconn = projectname;

		opened = UNCHANGED;
		guidstat = CLEAN;
	    CComPtr<IMgaTerritory> lm;
	    COMTHROW(CreateTerritory(NULL, &lm));
		COMTHROW(BeginTransaction(lm, TRANSACTION_READ_ONLY));

		try {	

			CoreObj  dataroot;
			COMTHROW(dataproject->get_RootObject(&dataroot.ComPtr()));
		    s=dataroot[ATTRID_PARADIGM];
			mgaversion = dataroot[ATTRID_MGAVERSION];

			pGUID=CComVariant(dataroot[ATTRID_PARGUID]);
			ver = dataroot[ATTRID_PARVERSION];
			COMTHROW(lm->Flush());
			COMTHROW(CommitTransaction());
		} catch(hresult_exception &e) {
			lm->Flush();
			AbortTransaction();
			throw e;
		}

		CComBSTR soldname = s;
		CComVariant soldguid = pGUID;

		if(paradigmname) {
			ver.Empty();
			s = paradigmname;
		}
		if(paradigmGUID.vt != VT_EMPTY) {
			ver.Empty();
			if(CComVariant(true) == paradigmGUID) pGUID = NULLVARIANT;
			else pGUID = paradigmGUID;
		}
		if(s.Length()) {
			if (ver.Length()) {
				// Version string has precedence
				COMTHROW(OpenParadigm(s,ver));
				COMTHROW(metapr->get_GUID(&pGUID));
			}
			else {
				COMTHROW(OpenParadigm(s,&pGUID));
				COMTHROW(metapr->get_Version(&ver));
			}
			bool guidchanged = false;
			if(guidcmp(soldguid, pGUID)) {
				COMTHROW(BeginTransaction(lm, TRANSACTION_READ_ONLY));
				try {
					CComPtr<IMgaFolder> rf;
					COMTHROW(get_RootFolder(&rf));
					COMTHROW(ObjFor(rf)->CheckTree());
					COMTHROW(CommitTransaction());
				} catch(hresult_exception &e) {
					lm->Flush();
					AbortTransaction();
					throw e;
				}
				guidchanged = true;
			}
			if(s != soldname || guidchanged) {
				if(ro) COMTHROW(E_MGA_READ_ONLY_ACCESS);
				COMTHROW(BeginTransaction(lm, TRANSACTION_GENERAL));
				try {
					CoreObj  dataroot;
					COMTHROW(dataproject->get_RootObject(&dataroot.ComPtr()));
					dataroot[ATTRID_PARADIGM] = s;
					dataroot[ATTRID_PARGUID] = pGUID;
					dataroot[ATTRID_PARVERSION] = ver;
					COMTHROW(CommitTransaction());
					opened = CHANGED;
					guidstat = DIRTY;
				} catch(hresult_exception &e) {
					lm->Flush();
					AbortTransaction();
					throw e;
				}
			}
		}
		else COMTHROW(E_INVALID_USAGE);
		COMTHROW(RegisterActiveObject((IMgaProject *)this,CLSID_MgaProject,ACTIVEOBJECT_STRONG,&rot));
		StartAutoAddOns();
		try {
			COMTHROW(BeginTransaction(lm, TRANSACTION_READ_ONLY));
			GlobalNotify(GLOBALEVENT_OPEN_PROJECT);
			COMTHROW(lm->Flush());
			COMTHROW(CommitTransaction());
		} catch(hresult_exception &e) {
			lm->Flush();
			AbortTransaction();
			throw e;
		}
		MARKSIG('2');
	} 
	COMCATCH(
		opened = CLOSED;
		guidstat = CLEAN;
		if(dataproject) dataproject->CloseProject();
		if(metapr) metapr->Close();
		metapr = 0;
		projconn.Empty(); 
		parconn.Empty();
	)
}

STDMETHODIMP CMgaProject::Open(BSTR projectname, VARIANT_BOOL *ro_mode)
{
	COMTRY {
		if(opened != CLOSED) COMTHROW(E_MGA_PROJECT_OPEN);
		CComBSTR s;
		CComVariant pGUID;
		CComBSTR ver;
		CComPtr<ICoreMetaProject> genericproject;
		CreateCoreMetaProject(genericproject);

		COMTHROW(dataproject->OpenProject(projectname, genericproject, ro_mode));
		
		projconn = projectname;

		opened = UNCHANGED;
		guidstat = CLEAN;
	    CComPtr<IMgaTerritory> lm;
	    COMTHROW(CreateTerritory(NULL, &lm));
		COMTHROW(BeginTransaction(lm, TRANSACTION_READ_ONLY));

		try {	

			CoreObj  dataroot;
			COMTHROW(dataproject->get_RootObject(&dataroot.ComPtr()));
		    s=dataroot[ATTRID_PARADIGM];
			mgaversion = dataroot[ATTRID_MGAVERSION];

			pGUID=CComVariant(dataroot[ATTRID_PARGUID]);
			ver=dataroot[ATTRID_PARVERSION];
			COMTHROW(lm->Flush());
			COMTHROW(CommitTransaction());
		} catch(hresult_exception &e) {
			lm->Flush();
			AbortTransaction();
			throw e;
		}
	
		if(s.Length()) {
			if (ver.Length()) {
				// Version string has precedence
				COMTHROW(OpenParadigm(s,ver));
			}
			else {
				COMTHROW(OpenParadigm(s,&pGUID));
			}
		}
		else COMTHROW(E_MGA_MODULE_INCOMPATIBILITY);

		CComVariant nGUID;
		COMTHROW(metapr->get_GUID(&nGUID));

		if(guidcmp(pGUID, nGUID)) {
			COMTHROW(BeginTransaction(lm, TRANSACTION_READ_ONLY));
			try {
				CComPtr<IMgaFolder> rf;
				COMTHROW(get_RootFolder(&rf));
				COMTHROW(ObjFor(rf)->CheckTree());
				COMTHROW(CommitTransaction());
			} catch(hresult_exception &e) {
				lm->Flush();
				AbortTransaction();
				throw e;
			}

			if (*ro_mode == VARIANT_FALSE) {
				COMTHROW(BeginTransaction(lm, TRANSACTION_GENERAL));
				try {
					CoreObj  dataroot;
					COMTHROW(dataproject->get_RootObject(&dataroot.ComPtr()));
					dataroot[ATTRID_PARGUID] = nGUID;
					COMTHROW(CommitTransaction());
					opened = CHANGED;
					guidstat = DIRTY;
				} catch(hresult_exception &e) {
					lm->Flush();
					AbortTransaction();
					throw e;
				}
			}
		}

		COMTHROW(RegisterActiveObject((IMgaProject *)this,CLSID_MgaProject,ACTIVEOBJECT_STRONG,&rot));
		StartAutoAddOns();
	
		try {
			COMTHROW(BeginTransaction(lm, TRANSACTION_READ_ONLY));
			GlobalNotify(GLOBALEVENT_OPEN_PROJECT);
			COMTHROW(CommitTransaction());
		} catch(hresult_exception &e) {
			lm->Flush();
			AbortTransaction();
			throw e;
		}

		MARKSIG('2');
	} 
	COMCATCH(
		opened = CLOSED;
		if(dataproject) dataproject->CloseProject();
		if(metapr) metapr->Close();
		metapr = 0;
		projconn.Empty(); 
		parconn.Empty();
	)
}

STDMETHODIMP CMgaProject::QueryProjectInfo(BSTR projectname, long *mgaversion, 
										   BSTR *paradigmname, 
										   BSTR *parversion,
										   VARIANT *paradigmGUID,
										   VARIANT_BOOL *ro_mode)
{
	CComPtr<ICoreProject> dp;
	COMTRY {
		CHECK_INSTRPAR(projectname);
		CHECK_OUTPAR(mgaversion);
		CHECK_OUTSTRPAR(paradigmname);
		CHECK_OUTSTRPAR(parversion);
		CHECK_OUTPAR(paradigmGUID);
		CHECK_OUTPAR(ro_mode);

		CComPtr<ICoreMetaProject> genericproject;
		CreateCoreMetaProject(genericproject);

		COMTHROW(dp.CoCreateInstance(CLSID_CoreProject));
		COMTHROW(dp->OpenProject(projectname, genericproject, ro_mode));
		
		CComPtr<ICoreTerritory> tt;
		COMTHROW(dp->CreateTerritory(&tt));
		COMTHROW(dp->BeginTransaction(TRANSTYPE_READFIRST));

		try {	
			COMTHROW(dp->PushTerritory(tt));

			CoreObj  dataroot;
			COMTHROW(dp->get_RootObject(&dataroot.ComPtr()));
		    *paradigmname = CComBSTR(dataroot[ATTRID_PARADIGM]).Detach();
		    *parversion = CComBSTR(dataroot[ATTRID_PARVERSION]).Detach();
		    CComVariant(dataroot[ATTRID_PARGUID]).Detach(paradigmGUID);
			*mgaversion = dataroot[ATTRID_MGAVERSION];
			dp->PopTerritory();
		} catch(hresult_exception &) {
			;
		}
		dp->AbortTransaction();
		dp->CloseProject();
	}
	COMCATCH(
		if(dp) dp->CloseProject();
	)
}

STDMETHODIMP CMgaProject::Save(BSTR newname, VARIANT_BOOL keepoldname)
{
	COMTRY {
		if(baseterr) COMTHROW(E_MGA_ALREADY_IN_TRANSACTION);
		{
			CComPtr<IMgaTerritory> t;
			COMTHROW(CreateTerritory(NULL, &t));
			if(opened >= CHANGED) {
				COMTHROW(BeginTransaction(t, TRANSACTION_GENERAL));
				try {
					CoreObj self;
					COMTHROW(dataproject->get_RootObject(&self.ComPtr()));
					FixupGUID();
					locking_type ppp;
					COMTHROW(self->get_PeerLockValue(ATTRID_LOCK, &ppp));
					//if(!(ppp & LOCKING_READ)) self[ATTRID_MDATE] = Now();
                    // BGY: this modifies the root every save time, which is bad for the
                    // xmlbackend (it always tries to check out the root)
					COMTHROW(CommitTransaction());
				} catch(hresult_exception &) {
					t->Flush();
					AbortTransaction();
				}
			}
		}
		COMTHROW(dataproject->SaveProject(newname, keepoldname));
		if(CComBSTR(newname).Length()) {
			if (!keepoldname) {
				projconn = newname;
				opened = UNCHANGED;
				transactioncount = 0;
			}

		} else {
			opened = UNCHANGED;
			transactioncount = 0;
		}
	}
	COMCATCH(;);
}

STDMETHODIMP CMgaProject::Close(VARIANT_BOOL abort)
{
	if(opened == CLOSED) {
		ASSERT(("Project is closed but transaction is active", !baseterr));
		return S_OK;
	}
	if(baseterr) COMTHROW(AbortTransaction());
	
	COMTRY {
		{
			CComPtr<IMgaTerritory> t;
			COMTHROW(CreateTerritory(NULL, &t));
			bool write = !abort && opened >= CHANGED;
			COMTHROW(BeginTransaction(t, write ? TRANSACTION_GENERAL : TRANSACTION_READ_ONLY));
			try {
				if(write) {
					CoreObj self;
					COMTHROW(dataproject->get_RootObject(&self.ComPtr()));
					FixupGUID();
					locking_type ppp;
					COMTHROW(self->get_PeerLockValue(ATTRID_LOCK, &ppp));
					//if(!(ppp & LOCKING_READ)) self[ATTRID_MDATE] = Now();
                    // BGY: this modifies the root every closing time, which is bad for the
                    // xmlbackend (it always tries to check out the root)
				}
				GlobalNotify(GLOBALEVENT_CLOSE_PROJECT);
				COMTHROW(CommitTransaction());
			} catch(hresult_exception &) {
				t->Flush();
				AbortTransaction();
			}
		}
		
		StopAutoAddOns();  // PETER: Moved these two lines here, otherwise addons won't receive notifications
		RemoveClients();   //

	 	if(rot) COMTHROW(RevokeActiveObject(rot,NULL));
		if(dataproject) COMTHROW(dataproject->CloseProject(abort));

		if(metapr) {
			COMTHROW(metapr->Close());
			metapr = 0;
		}

	    MARKSIG('8');
		opened = CLOSED;
		guidstat = CLEAN;
		projconn.Empty(); 
		parconn.Empty();
		transactioncount = 0;
	}
	COMCATCH( opened = CLOSEERROR;);	//  You cannot rollback a failed Close completely, so I did not even try it.
}

#undef mgaproject


// ----------------------------------------
// Access the root nodes of data and meta
// ----------------------------------------
STDMETHODIMP CMgaProject::get_RootFolder(IMgaFolder **pVal)
{
	COMTRY {
		CHECK_OUTPTRPAR(pVal);
		CoreObj  dataroot;
		COMTHROW(dataproject->get_RootObject(&dataroot.ComPtr()));
		CoreObjs s = dataroot[ATTRID_FPARENT + ATTRID_COLLECTION];
		ITERATE_THROUGH(s) {
			CComPtr<ICoreMetaObject> m;
			metaid_type t;
			ITER->get_MetaObject(&m);
			m->get_MetaID(&t);
			if(t == DTID_FOLDER) {
				ObjForCore(ITER)->getinterface(pVal);
				break;
			}
		}
		if(*pVal == NULL) COMTHROW(E_MGA_MODULE_INCOMPATIBILITY);
    }
    COMCATCH(;);
}


STDMETHODIMP CMgaProject::get_RootMeta(IMgaMetaProject **pVal)
{
	COMTRY {
		if(metapr == 0) COMTHROW( E_MGA_PROJECT_NOT_OPEN );
		*pVal = metapr; (*pVal)->AddRef();
	} COMCATCH(;);
}

// ----------------------------------------
// Filter and search functions
// ----------------------------------------
HRESULT CMgaProject::CreateFilter(IMgaFilter **newfilter) {
	COMTRY {
		CHECK_OUTPTRPAR(newfilter);
		CComPtr<CMgaFilter> filter;
		CreateComObject(filter);
		filter->setproject(this);
		*newfilter = filter.Detach();
	} COMCATCH(;);
}

void recursefolders(CoreObj folder, CMgaFilter * filter, EXCOLLECTIONTYPE_FOR(MgaFCO) *q) {
	{
		ITERATE_THROUGH(folder[ATTRID_FCOPARENT+ATTRID_COLLECTION]) {
			if(ITER.IsFCO()) filter->searchFCOs(ITER, q);
			else recursefolders(ITER, filter, q);
		}
	}
}

STDMETHODIMP CMgaProject::AllFCOs(IMgaFilter *filter, IMgaFCOs ** fcos) {
	COMTRY {
		CHECK_OUTPTRPAR(fcos);
		CHECK_INPTRPAR(filter);   // project is checked in the following lines
		CComPtr<IMgaProject> p;
		COMTHROW(filter->get_Project(&p));
		if(!COM_EQUAL(p, (IMgaProject *)this)) COMTHROW(E_MGA_FOREIGN_PROJECT);

		CREATEEXCOLLECTION_FOR(MgaFCO, q);
		CComPtr<IMgaFolder> rootf;
		get_RootFolder(&rootf);
		recursefolders(CoreObj(rootf), reinterpret_cast<CMgaFilter *>(filter),  q);
		*fcos = q.Detach();
    }
    COMCATCH(;);
}

STDMETHODIMP CMgaProject::GetFCOByID(BSTR id, IMgaFCO **pVal) {
	COMTRY {
		CComPtr<IMgaObject> p;
		COMTHROW(GetObjectByID(id,&p));
		COMTHROW(p.QueryInterface(pVal));
    } COMCATCH(;);
}

STDMETHODIMP CMgaProject::GetObjectByID(BSTR id, IMgaObject **pVal)
{

	COMTRY {
		CHECK_INSTRPAR(id);
		CHECK_OUTPTRPAR(pVal);
		metaid_type mm; // short
		objid_type ss;  // long
		if( swscanf(id,OLESTR("id-%04hx-%08lx"), &mm, &ss) != 2 || 
			mm < DTID_MODEL || mm > DTID_FOLDER) COMTHROW(E_MGA_BAD_ID);
		CoreObj obj;
		COMTHROW(dataproject->get_Object(mm,ss,&obj.ComPtr()));
		if(obj) ObjForCore(obj)->getinterface(pVal);
		else COMTHROW(E_MGA_BAD_ID);
    } COMCATCH(;);
}

STDMETHODIMP CMgaProject::GetFCOsByName(BSTR name, IMgaFCOs **pVal)
{

	COMTRY {
// ADD CODE
//				(*pVal)->Open()
		COMTHROW(E_MGA_NOT_IMPLEMENTED);
    }
    COMCATCH(;);
}

STDMETHODIMP CMgaProject::GetFolderByPath(BSTR path, IMgaFolder **pVal)
{

	COMTRY {
// ADD CODE
//				(*pVal)->Open()
		COMTHROW(E_MGA_NOT_IMPLEMENTED);
    }
    COMCATCH(;);
}

// see FCO::get_ObjectByPath for details in MgaFolder.cpp 
STDMETHODIMP CMgaProject::get_ObjectByPath(BSTR path, IMgaObject **pVal) {
	COMTRY {
		CHECK_INSTRPAR(path);
		CHECK_OUTPTRPAR(pVal);
		CComPtr<IMgaFolder> f;
		COMTHROW(get_RootFolder(&f));
		COMTHROW(f->get_ObjectByPath(path, pVal));
	} COMCATCH(;);
}

STDMETHODIMP CMgaProject::get_NthObjectByPath(long n_th, BSTR path, IMgaObject **pVal) {
	COMTRY {
		CHECK_INSTRPAR(path);
		CHECK_OUTPTRPAR(pVal);
		CComPtr<IMgaFolder> f;
		COMTHROW(get_RootFolder(&f));
		COMTHROW(f->get_NthObjectByPath(n_th, path, pVal));
	} COMCATCH(;);
}

// ----------------------------------------
// Access project properties
// ----------------------------------------
STDMETHODIMP CMgaProject::GetStatistics(BSTR *statstr)
{

	COMTRY {
// ADD CODE
    }
    COMCATCH(;);
}



STDMETHODIMP CMgaProject::get_CreateTime(BSTR *pVal)
{

	COMTRY {
		CoreObj self;
		COMTHROW(dataproject->get_RootObject(&self.ComPtr()));
		*pVal = CComBSTR(self[ATTRID_CDATE]).Detach();
	}
    COMCATCH(;);
}

STDMETHODIMP CMgaProject::get_ChangeTime(BSTR *pVal)
{

	COMTRY {
		CoreObj self;
		COMTHROW(dataproject->get_RootObject(&self.ComPtr()));
		*pVal = CComBSTR(self[ATTRID_MDATE]).Detach();
    }
    COMCATCH(;);
}

STDMETHODIMP CMgaProject::get_Author(BSTR *pVal)
{

	COMTRY {
		CoreObj self;
		COMTHROW(dataproject->get_RootObject(&self.ComPtr()));
		*pVal = CComBSTR(self[ATTRID_CREATOR]).Detach();
    }
    COMCATCH(;);
}

STDMETHODIMP CMgaProject::put_Author(BSTR newVal)
{

	COMTRY_IN_TRANSACTION {
		CoreObj self;
		COMTHROW(dataproject->get_RootObject(&self.ComPtr()));
		self[ATTRID_CREATOR] = newVal;
		notifyqueueprocessed = true;
		COMTHROW(GlobalNotify(GLOBALEVENT_PROJECT_PROPERTIES));
    }
    COMCATCH_IN_TRANSACTION(;);
}

STDMETHODIMP CMgaProject::get_MetaGUID(VARIANT *pVal)
{

	COMTRY
	{
		CHECK_OUTPAR(pVal);
		CoreObj self;
		COMTHROW(dataproject->get_RootObject(&self.ComPtr()));
		CComVariant p = self[ATTRID_PARGUID];

		if( p.vt != (VT_UI1 | VT_ARRAY) || GetArrayLength(p) != sizeof(GUID) )
		{
			GUID guid;
			memset(&guid, 0, sizeof(GUID));

			CopyTo(guid, p);
		}
		p.Detach(pVal);
	}
	COMCATCH(;)
}

STDMETHODIMP CMgaProject::get_MetaName(BSTR *pVal)
{

	COMTRY
	{
		CHECK_OUTSTRPAR(pVal);
		CoreObj self;
		COMTHROW(dataproject->get_RootObject(&self.ComPtr()));
		CComBSTR p = self[ATTRID_PARADIGM];

		*pVal=p.Detach();
	}
	COMCATCH(;)
}

STDMETHODIMP CMgaProject::get_MetaVersion(BSTR *pVal)
{

	COMTRY
	{
		CHECK_OUTSTRPAR(pVal);
		CoreObj self;
		COMTHROW(dataproject->get_RootObject(&self.ComPtr()));
		CComBSTR p = self[ATTRID_PARVERSION];

		*pVal=p.Detach();
	}
	COMCATCH(;)
}

STDMETHODIMP CMgaProject::get_GUID(VARIANT *pVal)
{

	COMTRY
	{
		CHECK_OUTPAR(pVal);
		
		FixupGUID(false);

		CoreObj self;
		COMTHROW(dataproject->get_RootObject(&self.ComPtr()));
		CComVariant p = self[ATTRID_GUID];

		
		if (guidstat == PENDING) {
			p = pendingguid;
		}
		
		if( p.vt != (VT_UI1 | VT_ARRAY) || GetArrayLength(p) != sizeof(GUID) )
		{
			GUID guid;
			memset(&guid, 0, sizeof(GUID));

			CopyTo(guid, p);
		}
		p.Detach(pVal);
	}
	COMCATCH(;)
}


STDMETHODIMP CMgaProject::put_GUID(VARIANT newVal)
{
	COMTRY_IN_TRANSACTION {
		CHECK_INPAR(newVal);
		CoreObj self;
		COMTHROW(dataproject->get_RootObject(&self.ComPtr()));

		if( newVal.vt != (VT_UI1 | VT_ARRAY) || GetArrayLength(newVal) != sizeof(GUID) )
			COMTHROW(E_INVALIDARG);

		self[ATTRID_GUID] = newVal;
		guidstat = MANUAL;
		notifyqueueprocessed = true;
		COMTHROW(GlobalNotify(GLOBALEVENT_PROJECT_PROPERTIES));
    }
    COMCATCH_IN_TRANSACTION(;);
}


STDMETHODIMP CMgaProject::get_Comment(BSTR *pVal)
{

	COMTRY {
		CoreObj self;
		COMTHROW(dataproject->get_RootObject(&self.ComPtr()));
		*pVal = CComBSTR(self[ATTRID_EXTDATA]).Detach();
    }
    COMCATCH(;);
}

STDMETHODIMP CMgaProject::put_Comment(BSTR newVal)
{

	COMTRY_IN_TRANSACTION {
		CoreObj self;
		COMTHROW(dataproject->get_RootObject(&self.ComPtr()));
		self[ATTRID_EXTDATA] = newVal;
		notifyqueueprocessed = true;
		COMTHROW(GlobalNotify(GLOBALEVENT_PROJECT_PROPERTIES));
    }
    COMCATCH_IN_TRANSACTION(;);
}

STDMETHODIMP CMgaProject::put_Name(BSTR newVal)
{

	COMTRY_IN_TRANSACTION {
		CHECK_INSTRPAR(newVal);
		CoreObj self;
		COMTHROW(dataproject->get_RootObject(&self.ComPtr()));
		self[ATTRID_NAME] = newVal;
		notifyqueueprocessed = true;
		COMTHROW(GlobalNotify(GLOBALEVENT_PROJECT_PROPERTIES));
    }
    COMCATCH_IN_TRANSACTION(;);
}

STDMETHODIMP CMgaProject::put_Version(BSTR newVal)
{

	COMTRY_IN_TRANSACTION {
		CHECK_INSTRPAR(newVal);
		CoreObj self;
		COMTHROW(dataproject->get_RootObject(&self.ComPtr()));
		self[ATTRID_VERSION] = newVal;
		notifyqueueprocessed = true;
		COMTHROW(GlobalNotify(GLOBALEVENT_PROJECT_PROPERTIES));
    }
    COMCATCH_IN_TRANSACTION(;);
}


STDMETHODIMP CMgaProject::get_Name(BSTR *pVal)
{

	COMTRY {
		CHECK_OUTSTRPAR(pVal);
		CoreObj self;
		COMTHROW(dataproject->get_RootObject(&self.ComPtr()));
		CComBSTR s = self[ATTRID_NAME];
		*pVal = s.Detach();
    }
    COMCATCH(;);
}

STDMETHODIMP CMgaProject::get_Version(BSTR *pVal)
{

	COMTRY {
		CHECK_OUTSTRPAR(pVal);
		CoreObj self;
		COMTHROW(dataproject->get_RootObject(&self.ComPtr()));
		CComBSTR s = self[ATTRID_VERSION];
		*pVal = s.Detach();
    }
    COMCATCH(;);
}


STDMETHODIMP CMgaProject::get_ProjectConnStr(BSTR *pVal)
{
	COMTRY {
		CHECK_OUTSTRPAR(pVal);
		CComBSTR s = projconn;
		*pVal = s.Detach();
    }
    COMCATCH(;);
}

STDMETHODIMP CMgaProject::get_ParadigmConnStr(BSTR *pVal)
{
	COMTRY {
		CHECK_OUTSTRPAR(pVal);
		CComBSTR s = parconn;
		*pVal = s.Detach();
    }
    COMCATCH(;);
}


////////////////////////////////////////////////////////////////////////////////////////////
// ----------------------------------------
// TRANSACTION HANDLING & NOTIFICATION 
// ----------------------------------------////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////



STDMETHODIMP CMgaProject::CreateTerritory(IMgaEventSink *sink, IMgaTerritory **pp, IMgaEventSink *rwsink) {
	COMTRY {
		CHECK_OUTPTRPAR(pp);
// MUST NOT RETURN
		CComPtr<ICoreTerritory> tt;
		COMTHROW(dataproject->CreateTerritory(&tt)); // terge ?? 
// MUST SUCCEED
		CComPtr< CMgaTerritory > ster;
		CreateComObject(ster);
		ster->coreterr = tt.Detach();
		ster->mgaproject = this;; 
		allterrs.push_front(ster);
		ster->handler=sink;
		ster->rwhandler=rwsink;
		*pp = ster.Detach();
    }
    COMCATCH(;);
}

STDMETHODIMP CMgaProject::CreateAddOn(IMgaEventSink *sink, IMgaAddOn **pp) {
	COMTRY {
		CHECK_INPTRPAR(sink);
		CHECK_OUTPTRPAR(pp);
// MUST NOT RETURN

		
		if(!reserveterr) {
			CComPtr<IMgaTerritory> r;
			COMTHROW(CreateTerritory(NULL, &r));
			reserveterr = static_cast<CMgaTerritory *>(r.p);
		}
// MUST SUCCEED
		CComPtr< CMgaAddOn > saddon;
		CreateComObject(saddon);
		saddon->mgaproject = this;
		alladdons.push_front(saddon);
		saddon->handler=sink;
		if(inautoaddoncreate) {
			saddon->SetAutomatic();
		}
		*pp = saddon.Detach();
    }
    COMCATCH(;);

}



STDMETHODIMP CMgaProject::get_ActiveTerritory(IMgaTerritory **aterr) {
	COMTRY {
		CHECK_OUTPTRPAR(aterr);
		if((*aterr = baseterr) != NULL) (*aterr)->AddRef();
	} COMCATCH(;)
}

STDMETHODIMP CMgaProject::get_AddOns(IMgaAddOns **addons) {
	COMTRY {
		CHECK_OUTPTRPAR(addons);
		CREATECOLLECTION_FOR(IMgaAddOn, q);
		for(addoncoll::iterator j = alladdons.begin(); j != alladdons.end(); j++) {
			q->Add(*j);
		}
		*addons = q.Detach();
	} COMCATCH(;)
}

STDMETHODIMP CMgaProject::get_Clients(IMgaClients **clients) {
	COMTRY {
		CHECK_OUTPTRPAR(clients);
		CREATECOLLECTION_FOR(IMgaClient, q);
		for(clientcoll::iterator j = allclients.begin(); j != allclients.end(); j++) {
			q->Add(*j);
		}
		*clients = q.Detach();
	} COMCATCH(;)
}

STDMETHODIMP CMgaProject::RegisterClient(BSTR name, IDispatch *OLEServer, IMgaClient **pVal) {
	COMTRY {
		CHECK_INPTRPAR(OLEServer);
		CHECK_INSTRPAR(name);

		CComPtr< CMgaClient > client;
		CreateComObject(client);
		client->active = true;
		client->mgaproject = this;
		client->ole_server = OLEServer;
		client->name = name;
		allclients.push_front(client);
		*pVal = client.Detach();
		// TODO register only if not registered
    }
    COMCATCH(;);
}

STDMETHODIMP CMgaProject::GetClientByName(BSTR name, IMgaClient **pVal) {
	COMTRY {
		CHECK_INSTRPAR(name);
		for(clientcoll::iterator j = allclients.begin(); j != allclients.end(); j++) {
			CComPtr< CMgaClient > client(*j);
			if (client->name == name) {
				*pVal = client.Detach();
				break;
			}
		}
		if(j== allclients.end()) COMTHROW(E_MGA_NAME_NOT_FOUND);
	}
    COMCATCH(;);
}


STDMETHODIMP CMgaProject::get_Territories(IMgaTerritories **terrs) {
	COMTRY {
		CHECK_OUTPTRPAR(terrs);
		CREATECOLLECTION_FOR2(IMgaTerritory, IMgaTerritories, q);
		for(tercoll::iterator j = allterrs.begin(); j != allterrs.end(); j++) {
			q->Add(*j);
		}
		*terrs = q.Detach();
	} COMCATCH(;)
}

STDMETHODIMP CMgaProject::get_AddOnComponents(IMgaComponents **comps) {
	COMTRY {
		CHECK_OUTPTRPAR(comps);
		CREATECOLLECTION_FOR(IMgaComponent, q);
		for(compcoll::iterator j = autocomps.begin(); j != autocomps.end(); j++) {
			q->Add(*j);
		}
		*comps = q.Detach();
	} COMCATCH(;)
}

STDMETHODIMP CMgaProject::EnableAutoAddOns(VARIANT_BOOL bEnable) {
	COMTRY {
		if(baseterr) COMTHROW(E_MGA_ALREADY_IN_TRANSACTION);
		bool bnew = (bEnable == VARIANT_TRUE);
		if(bnew == autoaddons) return S_OK;
		autoaddons = bnew;
		if(opened != CLOSED) {
			if(autoaddons) StartAutoAddOns();
			else StopAutoAddOns();
		}
	} COMCATCH(;)
}


void CMgaProject::StartAutoAddOns() {
	if(!autoaddons) return;
#ifdef DEBUG
	for(addoncoll::iterator j = alladdons.begin(); j != alladdons.end(); j++) {
		ASSERT(!(*j)->IsAutomatic());
	}
#endif
	CComPtr<IMgaRegistrar> reg;
	CComBSTR paradigm;
	COMTHROW(metapr->get_Name(&paradigm));
	CComVariant progids;
	COMTHROW(reg.CoCreateInstance(OLESTR("MGA.MgaRegistrar")));
	COMTHROW(reg->get_AssociatedComponents(paradigm, COMPONENTTYPE_ADDON, REGACCESS_BOTH, &progids));
	long p = GetArrayLength(progids);
	ASSERT(p >= 0);
	if(p)  {
		std::vector<CComBstrObj> vec;
		vec.resize(p);
		CopyTo(progids, vec.begin(), vec.end());
		inautoaddoncreate = true;
		CComBSTR errs;
		for(std::vector<CComBstrObj>::iterator i = vec.begin(); i < vec.end(); i++) {
			try {
				CComPtr<IMgaComponent> addon;
				COMTHROW(CreateMgaComponent(addon, *i)); // Was: COMTHROW( addon.CoCreateInstance(*i) );
				ASSERT( addon != NULL );
				CComQIPtr<IMgaVersionInfo> vv=addon;
				MgaInterfaceVersion_enum v = MgaInterfaceVersion_None;
				if(vv) COMTHROW(vv->get_version(&v));
				if(v != MgaInterfaceVersion_Current) HR_THROW(E_MGA_COMPONENT_ERROR);

				COMTHROW( addon->Initialize(this));
				autocomps.push_front(addon.Detach());
			} catch(hresult_exception &) {
				errs += " ";
				errs += i->p;
			}
		}
		inautoaddoncreate = false;
		if(errs) {
				COMTHROW(E_MGA_COMPONENT_ERROR); // change error type
		}		
	}
//	SetAutomatic();
	
}

void CMgaProject::StopAutoAddOns() {
	while(!autocomps.empty()) {
			CComPtr<IMgaComponent> addon;
			addon.Attach(autocomps.front());
			autocomps.pop_front();
			addon.Release();
	}
#ifdef DEBUG
		for(addoncoll::iterator j = alladdons.begin(); j != alladdons.end();j++) {
			ASSERT(!(*j)->IsAutomatic());
		}
#endif
	CoFreeUnusedLibraries();
}

void CMgaProject::RemoveClients() {
	while(!allclients.empty()) {
			CComPtr< CMgaClient > client;
			client = allclients.front();
			client->active = false;
			allclients.pop_front();
	}
}


STDMETHODIMP CMgaProject::get_ProjectStatus(long *status) {
	COMTRY {
		CHECK_OUTPAR(status);
		if(opened == CLOSED) {
			*status  = 0;
		}	
		else if(opened == CLOSEERROR) {
			*status = 0x80000000;
		}
		else {
			*status = 1 + (opened == CHANGED ? 4 : 0) + (baseterr ? 8 : 0) + (read_only ? 16 : 0);
		}
	} COMCATCH(;);
}

STDMETHODIMP CMgaProject::BeginTransaction(IMgaTerritory *ter, transactiontype_enum mode)
{
	COMTRY {
		CComPtr<IMgaTerritory> ttemp;
		if(baseterr) COMTHROW(E_MGA_ALREADY_IN_TRANSACTION);
		if(!ter) {
			CreateTerritory(NULL,&ttemp);
			ter = ttemp;
		}
		ASSERT(mode == TRANSACTION_GENERAL || mode == TRANSACTION_READ_ONLY  || mode == TRANSACTION_NON_NESTED);
		CComPtr<IMgaProject> p;
		ter->get_Project(&p);
		if( p != this) COMTHROW(E_MGA_FOREIGN_OBJECT);
		CMgaTerritory *t = static_cast<CMgaTerritory *>(ter);
		if(!t->coreterr) COMTHROW(E_MGA_TARGET_DESTROYED);
		read_only = (mode == TRANSACTION_READ_ONLY);
		non_nestable = (mode == TRANSACTION_NON_NESTED);
		COMTHROW(dataproject->BeginTransaction(read_only? TRANSTYPE_READFIRST : TRANSTYPE_FIRST));
		checkofftemporary = false;
		in_nested = false;
		must_abort = false;
		COMTHROW(dataproject->PushTerritory(t->coreterr));
		baseterr = activeterr = t;
		notifyqueueprocessed = false;
		MARKSIG('3');
    }
    COMCATCH(;);
}

STDMETHODIMP CMgaProject::CommitTransaction()
{
	COMTRY {
		ASSERT(!in_nested);
		if(!baseterr) COMTHROW(E_MGA_NOT_IN_TRANSACTION);
		if(checkoff) COMTHROW(CheckSupress(VARIANT_FALSE));
		if(!notifyobjs.empty()) {
			CoreObj self;
			COMTHROW(dataproject->get_RootObject(&self.ComPtr()));
//			self[ATTRID_MDATE] = Now();
		}
		COMTHROW(CommitNotify());		
		COMTHROW(dataproject->PopTerritory());
		COMTHROW(dataproject->CommitTransaction(read_only ? TRANSTYPE_READFIRST: TRANSTYPE_FIRST));
        baseterr = activeterr= NULL;
		read_only = false;

		if(notifyqueueprocessed) {
			transactioncount++;
			opened = CHANGED;
			if (guidstat == MANUAL) {
				guidstat = CLEAN;
			}
			else {
				guidstat = DIRTY;
			}
		}
	    MARKSIG('7');
    }
    COMCATCH(;);
}



STDMETHODIMP CMgaProject::AbortTransaction() {
	COMTRY {
		ASSERT(!in_nested);
		if(!baseterr) COMTHROW(E_MGA_NOT_IN_TRANSACTION);
		aurcnt++;
		//COMTHROW(dataproject->PopTerritory()); BGY: see committransaction, it is already popped
		COMTHROW(dataproject->AbortTransaction(read_only ? TRANSTYPE_READFIRST:TRANSTYPE_FIRST));
		baseterr = activeterr= NULL;
		deferredchecks.clear();
		checkoff = false;
		while(!changedobjs.empty()) {
			changedobjs.front()->objforgetrwnotify();
			changedobjs.pop();
		}
		while(!notifyobjs.empty()) {
			notifyobjs.front()->objforgetnotify();
			notifyobjs.pop();
		}
		read_only = true;
		{
			CComPtr<IMgaTerritory> t;
			COMTHROW(CreateTerritory(NULL, &t));
			COMTHROW(BeginTransaction(t, TRANSACTION_READ_ONLY));
			GlobalNotify(GLOBALEVENT_ABORT_TRANSACTION);
			COMTHROW(CommitTransaction());
		}
		read_only = false;
	    MARKSIG('6');
    }
    COMCATCH(read_only = false;);
}


STDMETHODIMP CMgaProject::CheckSupress(VARIANT_BOOL s) {
	COMTRY_IN_TRANSACTION {
		if(s) checkoff = true;
		else { 
			for(objhash::iterator i = deferredchecks.begin(); i != deferredchecks.end(); i++) {
				if(GetMetaID((*i)->self) == DTID_FOLDER) COMTHROW((*i)->Check());
				else COMTHROW((*i)->CheckRCS());
			}
			deferredchecks.clear();
			checkoff = false;
		}
	} COMCATCH_IN_TRANSACTION(;);
}


STDMETHODIMP CMgaProject::GlobalNotify(globalevent_enum msg) {
  COMTRY {
		{
			addoncoll::iterator i = alladdons.begin(), end = alladdons.end();
			if(i != end) {
				COMTHROW(pushterr(*reserveterr));
// The addon may Destroy() itself or release a reference while being notified
// But other addons are expected to remain in the list
				while(i != end) {
					CComPtr<CMgaAddOn> addon(*i++); // set the iterator to the next addon (since it may be deleted)
					if(msg == GLOBALEVENT_NOTIFICATION_READY && addon->notified == false) continue;
					if(addon->handler->GlobalEvent(msg)!= S_OK) {
						ASSERT(("Global notification failed", false));
					}
				}
				COMTHROW(popterr());
			}
		}
		tercoll::iterator i = allterrs.begin(), end = allterrs.end();
		for(;i != end; i++) {
			if(msg == GLOBALEVENT_NOTIFICATION_READY && (*i)->notified == false) continue;
			else (*i)->notified = false;
			if((*i)->handler) {
				COMTHROW(pushterr(**i));
				(*i)->handler->GlobalEvent(msg); // Silently ignore failures
				COMTHROW(popterr());
			}
		}
  } COMCATCH(;);
}

STDMETHODIMP CMgaProject::Notify(globalevent_enum event) {
	COMTRY {
		if(baseterr) COMTHROW(E_MGA_ALREADY_IN_TRANSACTION);
		{
			CComPtr<IMgaTerritory> t;
			COMTHROW(CreateTerritory(NULL, &t));
			COMTHROW(BeginTransaction(t, TRANSACTION_READ_ONLY));
			GlobalNotify(event);
			COMTHROW(CommitTransaction());
		}
    }
    COMCATCH(;);
}


STDMETHODIMP CMgaProject::CommitNotify() {
  if(read_only) {
                ASSERT(changedobjs.empty() && notifyobjs.empty());
                return S_OK;
  }
  COMTRY {
                if(!baseterr) COMTHROW(E_MGA_NOT_IN_TRANSACTION);

                if(!changedobjs.empty()) notifyqueueprocessed = true;
                while(!changedobjs.empty()) {
                                FCOPtr f = changedobjs.front();
                                changedobjs.pop();

#if(1)
                                COMTHROW(f->objrwnotify());
#else
                                if(f->objrwnotify() != S_OK) {
                                        ASSERT(("Notification failed", false));
                                }
#endif
                }

                read_only = true;
                while(!notifyobjs.empty()) {
                                FCOPtr f = notifyobjs.front();
                                notifyobjs.pop();
                                if(f->objnotify() != S_OK) {
                                        ASSERT(("Notification failed", false));
                                }
                }
                COMTHROW(GlobalNotify(GLOBALEVENT_NOTIFICATION_READY));
                read_only = false;
  } COMCATCH( read_only = false;);
}



bool CMgaProject::alreadynested() { return in_nested; }

HRESULT CMgaProject::beginnested() {
	ASSERT(!in_nested);

	ASSERT(objstocheck.empty());
	objstocheck.clear();		

	HRESULT hr;
	if(non_nestable) hr = must_abort ? E_MGA_MUST_ABORT : S_OK;
	else hr = dataproject->BeginTransaction(TRANSTYPE_NESTED);

    MARKSIG('4');
	if(hr == 0) in_nested = true;
	return hr;
}

HRESULT CMgaProject::commitnested() {
	ASSERT(in_nested);

	ASSERT(objstocheck.empty());
	objstocheck.clear();

	HRESULT hr = S_OK;
	if(!non_nestable) hr = dataproject->CommitTransaction(TRANSTYPE_NESTED);
    MARKSIG('6');
	if(hr != S_OK) 	abortnested();
	else { 
		in_nested = false;
		while(!temporalobjs.empty()) {
			temporalobjs.front()->objrecordchange();
			temporalobjs.pop();
		}
	}
	return hr;
}

HRESULT CMgaProject::abortnested() {
	ASSERT(in_nested);
	objstocheck.clear();
	in_nested = false;
	while(!temporalobjs.empty()) {
			temporalobjs.front()->objforgetchange();
			temporalobjs.pop();
	}
    MARKSIG('5');
	must_abort = true;
	return dataproject->AbortTransaction(TRANSTYPE_NESTED);
}


HRESULT CMgaProject::pushterr(CMgaTerritory &ter) {
	COMTRY {
		ASSERT(("Territorys overwrite each other",activeterr==baseterr));
		activeterr = &ter;
		COMTHROW(dataproject->PushTerritory(ter.coreterr));
	} COMCATCH(;);
}

HRESULT CMgaProject::popterr() {
	COMTRY {
		activeterr= baseterr;
		COMTHROW(dataproject->PopTerritory());
	} COMCATCH(;);
}



STDMETHODIMP CMgaProject::Undo() {
	COMTRY {
		if(baseterr) COMTHROW(E_MGA_ALREADY_IN_TRANSACTION);
		COMTHROW(dataproject->UndoTransaction());
		if(!--transactioncount) {
			opened = UNCHANGED;
			guidstat = CLEAN;
		}
		{
			aurcnt++;
			CComPtr<IMgaTerritory> t;
			COMTHROW(CreateTerritory(NULL, &t));
			COMTHROW(BeginTransaction(t, TRANSACTION_READ_ONLY));
			GlobalNotify(GLOBALEVENT_UNDO);
			COMTHROW(CommitTransaction());
		}
    }
    COMCATCH(;);
}

STDMETHODIMP CMgaProject::Redo() {

	COMTRY {
		if(baseterr) COMTHROW(E_MGA_ALREADY_IN_TRANSACTION);
		COMTHROW(dataproject->RedoTransaction());
		transactioncount++;
		opened = CHANGED;
		guidstat = DIRTY;
		{
			aurcnt++;
			CComPtr<IMgaTerritory> t;
			COMTHROW(CreateTerritory(NULL, &t));
			COMTHROW(BeginTransaction(t, TRANSACTION_READ_ONLY));
			GlobalNotify(GLOBALEVENT_UNDO);
			COMTHROW(CommitTransaction());
		}
    }
    COMCATCH(;);
}

STDMETHODIMP CMgaProject::UndoRedoSize(short *undosize, short *redosize ) {
	COMTRY {
		COMTHROW(dataproject->get_UndoQueueSize(undosize));
		COMTHROW(dataproject->get_RedoQueueSize(redosize));
    }
    COMCATCH(;);
}


STDMETHODIMP CMgaProject::FlushUndoQueue() {

	if( dataproject == NULL )
		COMRETURN(E_FAIL);

	COMTRY {
		if(baseterr) COMTHROW(E_MGA_ALREADY_IN_TRANSACTION);
		COMTHROW( dataproject->FlushUndoQueue() );
	}
	COMCATCH(;)
}

STDMETHODIMP CMgaProject::UpdateSourceControlInfo()
{
    CComPtr<IMgaTerritory> lm;
	COMTHROW(CreateTerritory(NULL, &lm));
	COMTHROW(BeginTransaction(lm, TRANSACTION_GENERAL));

    // hack: core will recognize this action as a source control status update
    CoreObj  dataroot;
    COMTHROW(dataproject->get_RootObject(&dataroot.ComPtr()));
    dataroot[ATTRID_MDATE] = "UpdateSourceControlInfo";

    COMTHROW(CommitTransaction());

    return S_OK;
}

void CMgaProject::ObjMark(IMgaObject *s, long mask) {
	ObjFor(s)->SelfMark(mask);
}

void CMgaProject::FixupGUID(bool write) {
	if (guidstat == DIRTY) {
		GUID newGUID;
		COMTHROW(CoCreateGuid(&newGUID));
		CopyTo(newGUID, pendingguid);
		guidstat = PENDING;
	}
	if ((guidstat == PENDING) && write) {
		ASSERT(baseterr);
		CoreObj self;
		COMTHROW(dataproject->get_RootObject(&self.ComPtr()));
		self[ATTRID_GUID] = pendingguid;
		guidstat = CLEAN;
	}
}




STDMETHODIMP CMgaProject::CheckCollection(IMgaFCOs *coll) {
	MGACOLL_ITERATE(IMgaFCO, coll) {
		HRESULT s;
		if((s = MGACOLL_ITER->CheckProject(this)) != S_OK) return s;
	}
	MGACOLL_ITERATE_END;
	return S_OK;
}

// by ZolMol
STDMETHODIMP CMgaProject::CheckFolderCollection(IMgaFolders *coll) {
	MGACOLL_ITERATE(IMgaFolder, coll) {
		HRESULT s;
		if((s = MGACOLL_ITER->CheckProject(this)) != S_OK) return s;
	}
	MGACOLL_ITERATE_END;
	return S_OK;
}















// ----------------------------------------
// OBSOLETE stuff: Clearlocks
// ----------------------------------------

#include <set>
#include "../core/CoreUtilities.h"

bool std::less<metaobjidpair_type>::operator ()(const metaobjidpair_type &a,const metaobjidpair_type &b) const {
	if(a.metaid == b.metaid) return a.objid < b.objid;
	return a.metaid < b.metaid;
}

void ClearLocks(ICoreStorage * storage, std::set<metaobjidpair_type> &mset, short mi, long oi, bool clear) {
	
	metaobjidpair_type mm; 
	mm.metaid = mi, mm.objid = oi;

	if(mset.find(mm) != mset.end()) return;
	mset.insert(mm);
	CComPtr<ICoreMetaProject> mp;
	COMTHROW(storage->get_MetaProject(&mp));
	CComPtr<ICoreMetaObject> mo;
	COMTHROW(mp->get_Object(mi, &mo));
	CComPtr<ICoreMetaAttributes> atts;
	COMTHROW(mo->get_Attributes(&atts));
	long count = 0;								
	COMTHROW( atts->get_Count(&count) );		
	CComPtr<ICoreMetaAttribute> *arrptr, *arrend, *array = new CComPtr<ICoreMetaAttribute>[count]; 
	if(count > 0) { 
		COMTHROW( atts->GetAll((unsigned long)count, &(*array)) ); 
	} 
	arrend = array+count; 
	for(arrptr = array; arrptr != arrend; arrptr++)  {
		unsigned char t;
		COMTHROW((*arrptr)->get_ValueType(&t)); 
		switch(t) {
		case VALTYPE_LOCK:
				COMTHROW(storage->put_MetaID(mi));
				COMTHROW(storage->OpenObject(oi));
				COMTHROW(storage->put_MetaAttribute((*arrptr)));
				{
				CComVariant vv;
				COMTHROW(storage->get_AttributeValue(&vv));
				ASSERT(vv.vt == VT_I2);
				if(vv.iVal) {
					COMTHROW(storage->put_AttributeValue(CComVariant(0)));
				}
				}
				break;
		case VALTYPE_COLLECTION:
				COMTHROW(storage->put_MetaID(mi));
				COMTHROW(storage->OpenObject(oi));
				COMTHROW(storage->put_MetaAttribute((*arrptr)));
				{
					CComVariant v;
					COMTHROW(storage->get_AttributeValue(&v));
					metaobjidpair_type *i = NULL;
					metaobjidpair_type *e = NULL;
					GetArrayBounds(v, i, e);
					while( i != e )	{
		 				ClearLocks(storage, mset, (*i).metaid, (*i).objid, clear);
						i++;
					}
				}
		}
	}
	delete []array;
}



STDMETHODIMP CMgaProject::CheckLocks(BSTR filename, VARIANT_BOOL clearlocks) {
  CComPtr<ICoreStorage> storage;
  CComPtr<ICoreMetaProject> genericproject;
  COMTRY {
	CreateCoreMetaProject(genericproject);
	COMTHROW(storage.CoCreateInstance(	OLESTR("MGA.CoreRepository")));
	COMTHROW(storage->put_MetaProject(genericproject));
	COMTHROW(storage->OpenProject(filename, NULL));
	COMTHROW(storage->BeginTransaction());
	std::set<metaobjidpair_type> ccc;
	ClearLocks(storage, ccc, 1, 1, clearlocks ? true : false);
	COMTHROW(storage->CommitTransaction());
  } COMCATCH(;)
}



