#ifndef MGA_COREXMLFILE_H
#define MGA_COREXMLFILE_H

#include "resource.h"
#include <fstream>//fstream.h
#include <list>
#include <map>
#include <hash_map>
#include <set>
#include <xercesc/dom/dom.hpp>
#include "..\Mga\MgaGeneric.h"
#include "Ssauto.h"
#include "ccauto.h"

const metaobjidpair_type ROOT_ID_PAIR = {METAID_ROOT, OBJID_ROOT};

struct GUID_less
{
	bool operator()(const GUID& a, const GUID& b) const
	{
        if( a.Data1 < b.Data1 )
            return true;
        else if( a.Data1 > b.Data1 )
            return false;
        else
        {
            DWORD * adw = (DWORD*)&a;
            DWORD * bdw = (DWORD*)&b;
            if( *(adw+1) < *(bdw+1) )
                return true;
            else if( *(adw+1) > *(bdw+1) )
                return false;
            else if( *(adw+2) < *(bdw+2) )
                return true;
            else if( *(adw+2) > *(bdw+2) )
                return false;
            else if( *(adw+3) < *(bdw+3) )
                return true;
            else 
                return false;
        }
	}
};

class XmlObject;

typedef std::set<XmlObject*>                 XmlObjSet;
typedef XmlObjSet::iterator                  XmlObjSetIter;
typedef std::vector<XmlObject*>              XmlObjVec;
typedef XmlObjVec::iterator                  XmlObjVecIter;
typedef std::map<metaid_type,attrid_type>    ParentMap;
typedef std::map<GUID, XmlObject*,GUID_less> GUIDToXmlObjectMap;
typedef GUIDToXmlObjectMap::iterator         GUIDToXmlObjectMapIter;


////////////////////////////////////////////////////////////////////////////////
// XmlAttrBase class
////////////////////////////////////////////////////////////////////////////////
class XmlAttrBase
{
public:
    static XmlAttrBase * create      (valtype_type type);

                         XmlAttrBase ();
    virtual             ~XmlAttrBase ();


    virtual valtype_type getType     () const            = 0;
    virtual void         fromVariant (VARIANT p)         {}
    virtual void         toVariant   (VARIANT *p) const  {}
    virtual void         fromString  (const char * str)  {}
    virtual void         toString    (std::string& str) const {}
};

typedef std::map<attrid_type,XmlAttrBase*>   AttribMap;
typedef AttribMap::iterator                  AttribMapIter;

////////////////////////////////////////////////////////////////////////////////
// XmlAttrLong class
////////////////////////////////////////////////////////////////////////////////

class XmlAttrLong: public XmlAttrBase
{
public:
                         XmlAttrLong ();

    virtual valtype_type getType     () const;
    virtual void         fromVariant (VARIANT p);
	virtual void         toVariant   (VARIANT *p) const;
    virtual void         fromString  (const char * str);
    virtual void         toString    (std::string& str) const;
protected:
    long    m_value;
};

////////////////////////////////////////////////////////////////////////////////
// XmlAttrReal class
////////////////////////////////////////////////////////////////////////////////

class XmlAttrReal: public XmlAttrBase
{
public:
                         XmlAttrReal ();

    virtual valtype_type getType     () const;
    virtual void         fromVariant (VARIANT p);
	virtual void         toVariant   (VARIANT *p) const;
    virtual void         fromString  (const char * str);
    virtual void         toString    (std::string& str) const;
protected:
    double  m_value;
};

////////////////////////////////////////////////////////////////////////////////
// XmlAttrString class
////////////////////////////////////////////////////////////////////////////////

class XmlAttrString: public XmlAttrBase
{
public:
    virtual valtype_type getType     () const;
    virtual void         fromVariant (VARIANT p);
	virtual void         toVariant   (VARIANT *p) const;
    virtual void         fromString  (const char * str);
    virtual void         toString    (std::string& str) const;
protected:
    std::string  m_value;
};

////////////////////////////////////////////////////////////////////////////////
// XmlAttrBinary class
////////////////////////////////////////////////////////////////////////////////

class XmlAttrBinary: public XmlAttrBase
{
public:
    virtual valtype_type getType     () const;
    virtual void         fromVariant (VARIANT p);
	virtual void         toVariant   (VARIANT *p) const;
    virtual void         fromString  (const char * str);
    virtual void         toString    (std::string& str) const;
protected:
    bindata m_value;
};

////////////////////////////////////////////////////////////////////////////////
// XmlAttrLock class
////////////////////////////////////////////////////////////////////////////////

class XmlAttrLock: public XmlAttrBase
{
public:
                         XmlAttrLock ();

    virtual valtype_type getType     () const;
    virtual void         fromVariant (VARIANT p);
	virtual void         toVariant   (VARIANT *p) const;
    virtual void         fromString  (const char * str);
    virtual void         toString    (std::string& str) const;
protected:
    lockval_type    m_value;
};

////////////////////////////////////////////////////////////////////////////////
// XmlAttrPointer class
////////////////////////////////////////////////////////////////////////////////

class XmlAttrPointer: public XmlAttrBase
{
public:
                         XmlAttrPointer ();

    virtual valtype_type getType        () const;
protected:
    XmlObject * m_parent;

    friend class CCoreXmlFile;
};

////////////////////////////////////////////////////////////////////////////////
// XmlAttrCollection class
////////////////////////////////////////////////////////////////////////////////

class XmlAttrCollection: public XmlAttrBase
{
public:
    virtual valtype_type getType     () const;
protected:
    XmlObjSet   m_children;

    friend class CCoreXmlFile;
};


////////////////////////////////////////////////////////////////////////////////
// XmlObject class
////////////////////////////////////////////////////////////////////////////////

class XmlObject
{
public:
    enum AttrSet
    {
        ATTR_PRIMARY,
        ATTR_SECONDARY,
        ATTR_ALL        
    };

public:
                XmlObject           (ICoreMetaObject *metaobject, bool createAllAttributes );
               ~XmlObject           ();

    bool        isContainer         ();    
        
protected:
    void        createAttributes        (ICoreMetaObject *metaobject, int attrSet );
    void        deleteSecondaryAttribs  ();

protected:
    metaid_type        m_metaid;
    GUID               m_guid;
    int                m_index;            // index in the m_objects array
    AttribMap          m_attributes;
    bool               m_loaded;           // true if all attributes are loded (otherwise only pointer, collections and locks stored)
    bool               m_deleted;
    bool               m_modified;
    bool               m_sourceControlInfoRegNode;  
    CTime              m_lastWriteTime;

    friend class CCoreXmlFile;
};


struct UnresolvedPointer
{
    XmlObject   * m_object;             // whose attribute is this
    attrid_type   m_attrib;
    GUID          m_pointedObjGuid;
};

typedef std::vector<UnresolvedPointer>       UnresolvedPointerVec;
typedef UnresolvedPointerVec::iterator  UnresolvedPointerVecIt;

struct UndoItem
{
    UndoItem( XmlObject * object, attrid_type attrId, CComVariant value )
    {
        m_object = object;
        m_value  = value;
        m_attrId = attrId;
    };

    attrid_type   m_attrId;
    XmlObject *   m_object;
    CComVariant   m_value;
};

typedef std::map<XmlAttrBase*, UndoItem>  UndoMap;
typedef UndoMap::iterator            UndoMapIter;


class ATL_NO_VTABLE CCoreXmlFile : 
	public CComObjectRootEx<CComSingleThreadModel>,
	public IDispatchImpl<ICoreStorage, &IID_ICoreStorage, &LIBID_CORELib>,
	public ISupportErrorInfoImpl<&IID_ICoreStorage>,
	public CComCoClass<CCoreXmlFile, &CLSID_CoreBinFile>
{
public:
    enum SourceControl
    {
        SC_NONE,
        SC_SOURCESAFE,
        SC_CLEARCASE
    };

    enum CheckOutSate
    {
        CS_NOT_CHECKEDOUT,
        CS_CURRENT_USER,
        CS_OTHER_USER
    };

public:
	CCoreXmlFile();
	~CCoreXmlFile();

BEGIN_COM_MAP(CCoreXmlFile)
	COM_INTERFACE_ENTRY(ICoreStorage)
	COM_INTERFACE_ENTRY(IDispatch)
	COM_INTERFACE_ENTRY(ISupportErrorInfo)
END_COM_MAP()

DECLARE_REGISTRY_RESOURCEID(IDR_CoreXmlFile)

public:
	STDMETHODIMP get_MetaProject        (ICoreMetaProject **p);
	STDMETHODIMP put_MetaProject        (ICoreMetaProject *p);
	STDMETHODIMP get_MetaObject         (ICoreMetaObject **p);
	STDMETHODIMP put_MetaObject         (ICoreMetaObject *p);
	STDMETHODIMP get_MetaID             (metaid_type *p);
	STDMETHODIMP put_MetaID             (metaid_type p);
	STDMETHODIMP get_MetaAttribute      (ICoreMetaAttribute **p);
	STDMETHODIMP put_MetaAttribute      (ICoreMetaAttribute *p);
	STDMETHODIMP get_AttrID             (attrid_type *p);
	STDMETHODIMP put_AttrID             (attrid_type p);
	STDMETHODIMP get_AttributeValue     (VARIANT *p);
	STDMETHODIMP put_AttributeValue     (VARIANT p);
	STDMETHODIMP OpenObject             (objid_type objid);
	STDMETHODIMP CreateObject           (objid_type *objid);
	STDMETHODIMP CloseObject            ();
	STDMETHODIMP DeleteObject           ();
	STDMETHODIMP LockObject             ();
	STDMETHODIMP OpenProject            (BSTR connection, VARIANT_BOOL *ro_mode);
	STDMETHODIMP CreateProject          (BSTR connection);
    STDMETHODIMP SaveProject            (BSTR connection, VARIANT_BOOL keepoldname);
	STDMETHODIMP CloseProject           (VARIANT_BOOL abort = VARIANT_FALSE);
	STDMETHODIMP DeleteProject          ();
	STDMETHODIMP BeginTransaction       ();
	STDMETHODIMP CommitTransaction      ();
	STDMETHODIMP AbortTransaction       ();
	STDMETHODIMP get_StorageType        (long *p);

protected:
    void         fillParentMap          ();

    void         closeMetaProject       ();

    void         openMetaObject         ();

    void         closeMetaObject        ();

    void         openMetaAttribute      ();

    void         closeMetaAttribute     ();

    void         parseConnectionString  ( BSTR connection );

    void         setFileNames           ();

    void         getContainerFileName   (XmlObject * obj, std::string& str, bool fullpath=true);

    void         getContainerName       (XmlObject * obj, std::string& name, std::string& type);
    
    // graph operatrions
    void         clearAll               ();

    void         addObject              (XmlObject * obj);

    void         deleteObject           (XmlObject * obj);

    void         setPointer             (XmlObject * obj, attrid_type attribId, XmlObject * pointed);

    void         setPointer             (VARIANT p);
    
    void         updateCollections      ();

    void         resolvePointers        (UnresolvedPointerVec& pointers);

    void         createSourceControlInfoRegNode(XmlObject * obj);

    void         createSourceControlInfoRegNodes();

    // getting connections
    void         getPointer             (XmlAttrPointer * attr, VARIANT * p);

    void         getCollection          (XmlAttrCollection * attr, VARIANT * p);

    XmlObject *  getContainer           (XmlObject * obj);

    void         getContainedObjects    (XmlObject * obj, XmlObjVec& vec);

    // check out
    void         getCheckOutContainers  (XmlObjSet& objects, XmlObjSet& containers);

    void         getDependetContainers  (XmlObject * obj, XmlObjSet& containers, XmlObjSet& processedObjects);

    bool         checkOutFiles          (XmlObjSet& containers);

    // object pointer and id conversion
    XmlObject *  objectFromObjId        (metaobjidpair_type idpair);

    void         objIdFromObject        (XmlObject * obj, metaobjidpair_type& idpair);
    
    // serialization
    void         writeBinaryCache       ();

    bool         readBinaryCache        ();

    void         createProjectFile      ();

    void         readProjectFile        ();
    
    void         writeAll               ();
    
    void         writeXMLFile           (XmlObject * container);

    void         writeObject            (XmlObject * obj, FILE * file, bool container, const char * prefix, CTime lastWriteTime );

    bool         moveContainerToDisk    (XmlObject * container);

    void         fullReadContainer      (XmlObject * container);

    void         readXMLFile            (const char * fileName, UnresolvedPointerVec& pointers, bool fullLoad );

    void         readObject             (XERCES_CPP_NAMESPACE::DOMElement * e, UnresolvedPointerVec& pointers, 
                                         XmlObject * parent, bool fullLoad, CTime lastWriteTime );

    void         readAll                ( bool fullLoad );

    void         getLatestAndLoad       ();

    // source control
    bool         getSourceSafeUserInfo  ();

    void         createSourceSafeDatabase();

    void         openSourceSafeDatabase ();

    void         getSourceSafePath      (XmlObject * obj, std::string& str);

    bool         isContainerReadOnly    (XmlObject * obj);

    bool         isContinerCheckedOut   (XmlObject * obj);

    void         getSSCheckOutUser      (XmlObject * obj, std::string& user);

    void         checkOutContainer      (XmlObject * obj);

    void         addToSourceControl     (XmlObject * container);

    void         getLatestVersion       ();

    void         checkInAll             ();

    void         checkInAll             ( bool keepCheckedOut );

    bool         createClearCaseApp     ();
    
    bool         isPathUnderClearCase   ( const char * path );

    bool         isFileCheckedOutCC     ( const char * path );

    int          getCheckOutStateCC     ( const char * path );

    void         checkOutFileCC         ( const char * path );

    void         checkInFileCC          ( const char * path );

    void         addFileToCC            ( const char * path );

    void         addDirToCC             ( const char * path );

    void         getLatestVerCC         ( const char * path );

    void         showUsedFiles          ( XmlObjSet& containers );

    void         setSourceControlInfo   ( XmlObject * container, const char * scInfo );

    void         getSourceControlInfo   ( XmlObject * container, std::string& scInfo );

    void         updateSourceControlInfo( XmlObject * container );

    void         updateSourceControlInfo();

    void         dumpSourceControlInfo  ();

    bool         filesModifedByOthers   ();

protected:
    // meta project
    CComObjPtr<ICoreMetaProject>    m_metaProject;

    // meta object
    CComObjPtr<ICoreMetaObject>     m_metaObject;
    metaid_type                     m_metaObjectId;

    // meta attribute
    CComObjPtr<ICoreMetaAttribute>  m_metaAttribute;
	attrid_type                     m_metaAttributeId;
    valtype_type                    m_metaAttributeValType;

    ParentMap                       m_parentMap;
    
    // project
    bool                            m_opened;
    bool                            m_modified;
    std::string                     m_parentFolderPath;   // absolute path of the folder contains the project folder
    std::string                     m_folderPath;   // absolute path of the folder contains the project files
    std::string                     m_cacheFileName;
    std::string                     m_projectFileName;
    std::string                     m_projectName;
    bool                            m_inTransaction;
    bool                            m_savedOnce;

    // source control info
    int                             m_sourceControl;

    // sourcesafe
    std::string                     m_vssDatabaseStr;
    std::string                     m_vssParentPath;
    std::string                     m_vssPath;
    CComObjPtr<IVSSDatabase>        m_vssDatabase;
    std::string                     m_vssUser;
    std::string                     m_vssPassword;

    // clearcase
    std::string                     m_clearCaseString;
    CComObjPtr<IClearCase>          m_clearCase;
    CComObjPtr<IClearTool>          m_clearTool;

    // objects
    XmlObjVec                       m_objects;
    GUIDToXmlObjectMap              m_objectsByGUID;
    XmlObject                     * m_openedObject;
    XmlObject                     * m_root;

    // transaction handling
    UndoMap                         m_undoMap;
    XmlObjSet                       m_modifiedObjects;
    XmlObjVec                       m_createdObjects;
    XmlObjSet                       m_deletedObjects;
    //XmlObjSet                       m_checkOutContainers;
};

#endif//MGA_COREXMLFILE_H
